@@ -14,23 +14,6 @@
 // the forwarding of broadcasts.
 //
 
-#include <errno.h>
-#include <fcntl.h>
-#include <inttypes.h>
-#include <linux/if.h>
-#include <linux/if_tun.h>
-#include <stddef.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <sys/ioctl.h>
-#include <sys/stat.h>
-#include <sys/time.h>
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <time.h>
-#include <unistd.h>
-
 #include "htable.h"
 #include "queue.h"
 
@@ -568,9 +551,8 @@ static int parse_port(char *port,struct Allowed *into) {
 }
 
 //** IP address parsing utility
-// Capture an optional bits sub phrase [/<bits>]
+// Capture an optional bits sub phrase [/<bits>], zeroing the "/"
 static int parse_bits(char *bits,int max,struct Allowed *into) {
-    into->bits = max;
     if ( bits ) {
 	*(bits++) = 0;
 	int b;
@@ -586,59 +568,86 @@ static int parse_bits(char *bits,int max,struct Allowed *into) {
 // Parse a command line argument as a declaration of an allowed
 // remote into the given <addr>.
 // Return 0 if ok and 1 otherwise
-// Formats: <ipv4-address>[/<bits>][:<port>][=keyfile]
+// Formats: <ipv4-address or fqdn>[/<bits>][:<port>][=keyfile]
 // Formats: <ipv6-address>[/<bits>][=keyfile]
-// Formats: \[<ipv6-address>[/<bits>]\][:<port>][=keyfile]
-// Formats: hostname:port[=keyfile]
+// Formats: \[<ipv6-address or fqdn>[/<bits>]\][:<port>][=keyfile]
 static int parse_allowed(char *arg,struct Allowed *into) {
     static char buffer[10000];
+    char *start = buffer;
+    char *p = buffer;
     int n = strlen( arg );
     if ( n > 9000 ) {
 	return 1; // excessively large argument
     }
     strcpy( buffer, arg );
     into->source = arg;
-    char * keyfile = strchr( buffer, '=' );
-    if ( keyfile ) {
-	*(keyfile++) = 0;
-	into->psk.keyfile = keyfile;
+    int family = 0;
+    char e;
+    int bracket = 0;
+    if ( *p == '[' ) {
+	// ipv6 address or fqdn unto "/]"
+	family = AF_INET6;
+	for ( p++; *p && !strchr( "/]", *p ); p++ ) ;
+	start++;
+	bracket = 1;
+    } else {
+	for ( ; *p && !strchr( ".:", *p); p++ ) ;
+	if ( *p == ':' ) {
+	    // ipv6 address unto "/="
+	    family = AF_INET6;
+	    for ( ; *p && !strchr( "/=", *p); p++ ) ;
+	} else {
+	    // ipv4 address or fqdn unto ":/="
+	    family = AF_INET;
+	    for ( ; *p && !strchr( ":/=", *p); p++ ) ;
+	}
     }
-#define B(b) b, b+1, b+2, b+3
-    if ( sscanf( buffer, "%hhu.%hhu.%hhu.%hhu", B(into->addr.bytes) ) == 4 ) {
-#undef B
-	// ipv4 address
+    e = *p;
+    *p = 0;
+    struct addrinfo *result;
+    struct addrinfo hints = { 0 };
+    hints.ai_family = family;
+    int err = getaddrinfo( start, 0, &hints, &result );
+    if ( err ) {
+	fprintf( stderr, "Lookup %s error %s\n", start, strerror( -err ) );
+	return 1; // host failure
+    }
+    if ( family == AF_INET && result->ai_addrlen == 16 ) {
+        memcpy( into->addr.bytes, ((void*)result->ai_addr) + 4, 4 );
 	into->addr.width = 4;
-	if ( parse_port( strchr( buffer, ':' ), into ) ) {
+    } else {
+        memcpy( into->addr.bytes, result->ai_addr, result->ai_addrlen );
+	into->addr.width = result->ai_addrlen;
+    }
+    into->bits = ( family == AF_INET )? 32 : 128;
+    if ( e == '/' ) {
+	if ( parse_bits( p, into->bits, into ) ) {
+	    fprintf( stderr, "bad bits\n" );
+	    return 1; // Failure
+	}
+	if ( bracket ) {
+	    p = strchr( p+1, ']' );
+	    if ( p == 0 ) {
+		fprintf( stderr, "missing closing bracket\n" );
+		return 1;
+	    }
+	}
+	p++;
+	e = *p;
+    }
+    into->addr.width = ( family == AF_INET )? 4 : 16;
+    if ( e == ':' ) {
+	if ( parse_port( p, into ) ) {
 	    fprintf( stderr, "bad port\n" );
 	    return 1;
 	}
-	if ( parse_bits( strchr( buffer, '/' ), 32, into ) ) {
-	    fprintf( stderr, "bad bits\n" );
-	    return 1;
-	}
-	return 0;
+	p = strchr( p+1, '=' );
+	e = *p;
     }
-    // ipv6 address
-    char * address = buffer;
-    into->port = 0;
-    if ( *buffer == '[' ) {
-	// bracketed form, necessary for port
-	char *end = strchr( buffer, ']' );
-	if ( end == 0 ) {
-	    return 1; // bad argument
-	}
-	address++;
-	*(end++) = 0;
-	if ( *end == ':' && parse_port( end, into ) ) {
-	    return 1;
-	}
-    }
-    into->addr.width = 16;
-    if ( parse_bits( strchr( address, '/' ), 128, into ) ) {
-	return 1;
-    }
-    if ( inet_pton( AF_INET6, address, into->addr.bytes ) != 1 ) {
-	return 1; // Bad IPv6
+    if ( e == '=' ) {
+	*(p++) = 0;
+	into->psk.keyfile = malloc( strlen( p ) );
+	strcpy( into->psk.keyfile, p );
     }
     return 0;
 }
